import asyncio
import logging 

from ..core.base_client import BaseClient

logger = logging.getLogger(__name__)

class AppInfoClient(BaseClient):

    def __init__(self, url, session=None, identity=None):
        super(AppInfoClient, self).__init__(url, session, identity)

    async def __get_info(self):
        
        url = f"/atlas_engine/api/v1/info"

        result = await self.do_get(url)

        return result

    def get_info(self):

        async def run_loop():

            result = await self.__get_info()

            return result

        logger.info(f"Connection to atlas engine at url '{self._url}'.")
        logger.info(f"Get info of the atlas engine.")

        loop = asyncio.new_event_loop()

        task = run_loop()
        result = loop.run_until_complete(task)

        loop.close()

        return result

    async def __get_authority(self):

        url = f"/atlas_engine/api/v1/authority"

        result = await self.do_get(url)

        return result

    def get_authority(self):

        async def run_loop():
            result = await self.__get_authority()

            return result

        logger.info(f"Connection to atlas engine at url '{self._url}'.")
        logger.info(f"Get info of the authority of the atlas engine.")

        loop = asyncio.new_event_loop()

        task = run_loop()
        result = loop.run_until_complete(task)

        loop.close()

        return result


import asyncio
import inspect
import logging
import traceback

from ..core import BaseClient, LoopHelper
from .functional_error import FunctionalError


logger = logging.getLogger(__name__)


class ExternalTaskWorker(BaseClient):

    ADDITIONAL_LOCK_DURATION = (100 * 1000) - 100
    EXTEND_LOCK_TIMEOUT = (100 * 1000)
    
    def __init__(self, url, session, identity, loop_helper, handler, external_task, options={}):
        super(ExternalTaskWorker, self).__init__(url, session, identity)
        self._loop_helper = loop_helper
        self._handler = handler
        self._external_task = external_task
        self._payload = self._external_task.get('payload', {})
        self._options = options

    def __with_external_task_param(self):

        def is_handler_a_func(func):
            spec = inspect.getfullargspec(func)
            is_func = inspect.isroutine(func)

            result = (len(spec.args) == 2 and is_func)

            return result

        def is_handler_callable(caller):
            spec = inspect.getfullargspec(caller)
            is_func = inspect.isroutine(caller)

            result = (len(spec.args) == 3 and not is_func)

            return result

        return is_handler_a_func(self._handler) or is_handler_callable(self._handler)

    async def start(self):

        async def extend_lock_task_wrapper():
            worker_id = self._external_task['workerId']
            external_task_id = self._external_task['id']

            additional_duration = self._options.get("additional_lock_duration", ExternalTaskWorker.ADDITIONAL_LOCK_DURATION)

            await self.extend_lock(worker_id, external_task_id, additional_duration)

        extend_lock_timeout = self._options.get("extend_lock_timeout", ExternalTaskWorker.EXTEND_LOCK_TIMEOUT)

        options = {'delay': extend_lock_timeout}
        extend_lock = self._loop_helper.register_background_task(extend_lock_task_wrapper, **options)
        
        external_task_id = self._external_task['id']
        logger.info(f"starting external task '{external_task_id}'")

        try:
            result = None
            
            if asyncio.iscoroutinefunction(self._handler):
                if self.__with_external_task_param():
                    result = await self._handler(self._payload, self._external_task)
                else:
                    result = await self._handler(self._payload)
            else:
                if self.__with_external_task_param():
                    result = self._handler(self._payload, self._external_task)
                else:
                    result = self._handler(self._payload)

            await self.__finish_task_successfully(result)
        except FunctionalError as fe:
            logger.warning(f"Finish external task with functional error (code: {fe.get_code()}, message: {fe.get_message()})")
            await self.__finish_task_with_functional_errors(fe.get_code(), fe.get_message())

        except Exception as e:
            formatted_lines = traceback.format_exc().splitlines()
            logger.error(f"Finish external task with technial error ({str(e)} -> {formatted_lines})")
            await self.__finish_task_with_technical_errors(str(e), str(formatted_lines))
        finally:
            self._loop_helper.unregister_background_task(extend_lock, "extend_lock background task")
            logger.info(f"finish external task '{external_task_id}'")

    async def extend_lock(self, worker_id, external_task_id, additional_duration):
        path = f"/atlas_engine/api/v1/external_tasks/{external_task_id}/extend_lock"

        request = {
            "workerId": worker_id,
            "additionalDuration": additional_duration
        }

        await self.do_post(path, request)

    async def __finish_task_successfully(self, result):
        logger.info(f"finish task {self._external_task['id']} successfully.")
        path = f"/atlas_engine/api/v1/external_tasks/{self._external_task['id']}/finish"

        payload = {
            "workerId": self._external_task['workerId'],
            "result": result
        }

        result = await self.do_put(path, payload)
        logger.debug(f"finished task {self._external_task['id']} successfully.")

    async def __finish_task_with_functional_errors(self, error_code, error_message):
        logger.warn(f"finished external task_with functional errors '{self._external_task}', '{error_code}', '{error_message}'.")
        path = f"/atlas_engine/api/v1/external_tasks/{self._external_task['id']}/handle_bpmn_error"

        payload = {
            "workerId": self._external_task['workerId'],
            "bpmnError": {
                "errorCode": error_code,
                "errorMessage": error_message
            }
        }

        await self.do_put(path, payload)

    async def __finish_task_with_technical_errors(self, error_message, error_details):
        logger.warn(f"finished task with technical errors '{self._external_task}', '{error_message}', '{error_details}'.")
        path = f"/atlas_engine/api/v1/external_tasks/{self._external_task['id']}/handle_service_error"

        payload = {
            "workerId": self._external_task['workerId'],
            "error": {
                "errorMessage": error_message,
                "errorDetails": error_details
            }
        }

        await self.do_put(path, payload)




import asyncio
import pytest

from atlas_engine_client.process_control import ProcessControlClient

from tests.helper_aiohttp import http_redirect, CaseControlledTestServer

TIMEOUT = 1

@pytest.mark.asyncio
async def test_server_info(http_redirect):

    sample_response = """
    {"name": "@process-engine/process_engine_runtime", 
    "version": "8.5.0", 
    "description": "Standalone application that provides access to the .TS implementation of the ProcessEngine.", 
    "license": "MIT", 
    "homepage": "https://www.process-engine.io/"}
    """

    async with CaseControlledTestServer() as server:
        http_redirect.add_server('api.process-engine.io', 56100, server.port)
        client = ProcessControlClient('http://api.process-engine.io:56100', http_redirect.session)

        task = asyncio.ensure_future(client.get_serverinfo())
        request = await server.receive_request(timeout=TIMEOUT)

        server.send_response(request,  
            text=sample_response,
            content_type='application/json'
        )

        result = await asyncio.wait_for(task, TIMEOUT)
    
        assert(result['version'] == '8.5.0')

@pytest.mark.asyncio
async def test_start_process_instance(http_redirect):

    sample_response = """
        {"correlationId": "4d7e2e8c-9650-4dfe-a7b4-61df4c8a50f2", 
        "processInstanceId": "efc02a27-ba80-4d97-ac98-6c2446edb91b"}
    """

    async with CaseControlledTestServer() as server:
        http_redirect.add_server('api.process-engine.io', 56100, server.port)
        client = ProcessControlClient('http://api.process-engine.io:56100', http_redirect.session)

        task = asyncio.ensure_future(client.start_process_instance("sample_process"))
        request = await server.receive_request(timeout=TIMEOUT)

        server.send_response(request,  
            text=sample_response,
            content_type='application/json'
        )

        result = await asyncio.wait_for(task, TIMEOUT)
    
        assert(result['correlationId'] is not None)
        assert(result['processInstanceId'] is not None)




import asyncio
import logging 

from ..core import base_client

logger = logging.getLogger(__name__)

class ProcessInstanceClient(base_client.BaseClient):

    def __init__(self, url, session=None, identity=None):
        super(ProcessInstanceClient, self).__init__(url, session, identity)

    def terminate(self, process_instance_id):

        async def run_loop(process_instance_id):
            url = f"/atlas_engine/api/v1/process_instances/{process_instance_id}/terminate"

            result = await self.do_put(url, {})

            return result

        logger.info(f"Connection to process engine at url '{self._url}'.")
        logger.info(f"Terminate the process-instance '{process_instance_id}'.")

        loop = asyncio.new_event_loop()

        task = run_loop(process_instance_id)
        result = loop.run_until_complete(task)

        loop.close()

        return result

    def retry(self, process_instance_id):

        async def run_loop(process_instance_id):
            url = f"/atlas_engine/api/v1/process_instances/{process_instance_id}/retry"

            result = await self.do_put(url, {})

            return result

        logger.info(f"Retry process instance '{process_instance_id}' for atlas engine at url '{self._url}'.")

        loop = asyncio.new_event_loop()

        task = run_loop(process_instance_id)
        result = loop.run_until_complete(task)

        loop.close()

        return result




import asyncio
import pytest

from atlas_engine_client.core.base_client import BaseClient

from tests.helper_aiohttp import http_redirect, CaseControlledTestServer

TIMEOUT = 1

@pytest.mark.asyncio
async def test_server_info(http_redirect):

    sample_response = """
    {"name": "@process-engine/process_engine_runtime", 
    "version": "8.5.0"
    """

    async with CaseControlledTestServer() as server:
        http_redirect.add_server('api.process-engine.io', 56100, server.port)
        client = BaseClient('http://api.process-engine.io:56100', http_redirect.session)

        task = asyncio.ensure_future(client.get_serverinfo())
        request = await server.receive_request(timeout=TIMEOUT)

        server.send_response(request,  
            text=sample_response,
            content_type='application/json'
        )

        result = await asyncio.wait_for(task, TIMEOUT)
    
        assert(result['version'] == '8.5.0')



import asyncio
import aiohttp
import json
import logging

logger = logging.getLogger(__name__)

class BaseClient:

    def __init__(self, url, session=None, identity=None):
        self._url = url
        self._session = session

        if identity is None:
            self._identity = {"token": "ZHVtbXlfdG9rZW4="}
        else:
            self._identity = identity

    async def __aenter__(self):
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        await self.close() 

    async def close(self):
        if self._session:
            await self._session.close()

    async def do_get(self, path, options={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        async with aiohttp.ClientSession() as session:

            current_session = self._session if self._session else session

            async with current_session.get(request_url, headers=headers) as response:
                response.raise_for_status()
                if response.status == 200:
                    return await response.json()

    async def do_post(self, path, payload, options={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        logger.debug(f"post request to {request_url} with json payload {payload}")

        async with aiohttp.ClientSession() as session:

            current_session = self._session if self._session else session

            async with current_session.post(request_url, json=payload, headers=headers) as response:
                logger.debug(f"handle response {response.status}")
                response.raise_for_status()
                if response.status in [200, 201, 202]:
                    return await response.json()
                elif response.status == 204:
                    return ""
                else:
                    raise Exception(f"TODO: need a better error {response.status}")
    
    async def do_put(self, path, payload, options={}):
        headers = self.__get_default_headers()
        headers.update(options.get('headers', {}))
        headers.update(self.__get_auth_headers())

        request_url = f"{self._url}{path}"

        logger.debug(f"post request to {request_url} with json payload {payload}")

        async with aiohttp.ClientSession() as session:

            current_session = self._session if self._session else session

            async with current_session.put(request_url, json=payload, headers=headers) as response:
                logger.debug(f"handle response {response.status}")
                response.raise_for_status()
                if response.status in [200, 201, 202]:
                    return await response.json()
                elif response.status == 204:
                    return ""
                else:
                    raise Exception(f"TODO: need a better error {response.status}")
    
    async def get_serverinfo(self):
        return await self.do_get('/atlas_engine/api/v1/info')

    def __get_auth_headers(self):
        identity = self.__get_identity()
        token = identity['token']
        return {'Authorization': 'Bearer {}'.format(token)}

    def __get_default_headers(self):
        return {'Content-Type': 'application/json'}

    def __get_identity(self):
        identity = self._identity

        if callable(self._identity):
            identity = self._identity()

        return identity
            



import asyncio
import logging
import signal
import os

logger = logging.getLogger(__name__)

_DEFAULT_DELAY = 1

class LoopHelper:
    
    # TODO: mm - remove kwargs with better readable params
    def __init__(self, loop=asyncio.get_event_loop(), **kwargs):
        self._loop = loop
        self._tasks = []

        self.on_shutdown = kwargs.get('on_shutdown', self.__internal_on_shutdown)

    def create_task(self, task_callback):
        task = self._loop.create_task(task_callback())
        self._tasks.append(task)

    def register_delayed_task(self, task_func, **options):
        logger.info(f"create delayed tasks with options ({options}).")
        task = self._loop.create_task(self.__create_delayed_task(task_func, **options))
        self._tasks.append(task)

        return task

    def unregister_delayed_task(self, delayed_task, msg=""):
        return self.__unregister_task(delayed_task, msg)

    async def __create_delayed_task(self, task_func, **options):
        async def _worker(delay):
            try:
                await asyncio.sleep(delay)

                if asyncio.iscoroutinefunction(task_func):
                    logger.debug("running delayed job (async)")
                    await task_func()
                else:
                    logger.debug("running delayed job (sync)")
                    task_func()
                    
            except asyncio.CancelledError as ce:
                logger.debug(f"Cancel the task {ce}")


        delay = options.get('delay', _DEFAULT_DELAY)
        return await _worker(delay)

    def register_background_task(self, task_func, **options):
        logger.info(f"create background worker with options ({options}).")
        task = self._loop.create_task(self.__create_background_task(task_func, **options))
        self._tasks.append(task)
        return task

    def unregister_background_task(self, background_task, msg=""):
        return self.__unregister_task(background_task, msg)

    def __unregister_task(self, task, msg):
        can_unregister = True

        if self._tasks.index(task) >= 0:
            logger.info(f"cancel and unregister task: {msg}")
            self._tasks.remove(task)

            try:
                task.cancel()
                logger.info(f"cancelled task: {msg}")
            except asyncio.CancelledError as ce:
                logger.error(f"__unregister_task: {ce}")
                pass
        else:
            logger.warning("did'nt found task to unregister")
            can_unregister = False

        return can_unregister

    async def __create_background_task(self, task_func, **options):
        async def _task(delay):
            running = True

            while running:
                try:
                    if with_delay:
                        logger.debug(f"background worker delay for {delay}")
                        await asyncio.sleep(delay)
                    
                    if asyncio.iscoroutinefunction(task_func):
                        logger.debug("running background job (async)")
                        await task_func() 
                    else:
                        logger.debug("running background job (sync)")
                        task_func()

                except asyncio.CancelledError:
                    running = False

        delay = options.get('delay', _DEFAULT_DELAY)
        with_delay = True if delay > 0 else False

        return await _task(delay)

    def start(self):
        logger.info("Starting event loop.")
        try:
            self.__register_shutdown()
            self._loop.run_forever()
        except KeyboardInterrupt:
            self._loop.close()

    def stop(self):
        logger.info("Stopping event loop.")
        for task in self._tasks:
            try:
                task.cancel()
            except Exception as e:
                logger.warning(f"Task stopped with exception {e}")

        self._loop.stop()

    async def __internal_on_shutdown(self):
        logger.debug('only internal on_shutdown called')
        await asyncio.sleep(0)

    def __register_shutdown(self):
        async def shutdown(sig):
            logger.info(f"Received exit signal {sig.name}...")

            await self.on_shutdown()

            self.stop()

        signal_handler = lambda sig: asyncio.create_task(shutdown(sig))

        signals = ()

        if self.is_win():
            signals = (signal.SIGTERM, signal.SIGINT, signal.CTRL_C_EVENT)
        else:
            signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT, signal.SIGQUIT)

            for s in signals:
                self._loop.add_signal_handler(s, signal_handler, s)

    def is_win(self):
        return os.name == 'nt'



from atlas_engine_client.app_info import AppInfoClient
from atlas_engine_client.external_task import ClientWrapper
from atlas_engine_client.flow_node_instance import FlowNodeInstanceClient
from atlas_engine_client.notification import NotificationClient
from atlas_engine_client.process_instance import ProcessInstanceClient
from atlas_engine_client.process_definition import ProcessDefinitionClient
from atlas_engine_client.user_task import UserTaskClient


class ClientFactory:

    def __init__(self):
        pass

    def create_app_info_client(self, engine_url):
        return AppInfoClient(engine_url)

    def create_external_task_client(self, engine_url):
        return ClientWrapper(engine_url)

    def create_flow_node_instance_client(self, engine_url):
        return FlowNodeInstanceClient(engine_url)

    def create_notification_client(self, engine_url):
        return NotificationClient(engine_url)

    def create_process_instance_client(self, engine_url):
        return ProcessInstanceClient(engine_url)

    def create_process_definition_client(self, engine_url):
        return ProcessDefinitionClient(engine_url)

    def create_user_task_client(self, engine_url):
        return UserTaskClient(engine_url)

    



from ..core import base_client


class UserTaskClient(base_client.BaseClient):

    def __init__(self, url, session=None, identity=None):
        super(UserTaskClient, self).__init__(url, session, identity)

    async def reserve_user_task(self, user_task_instance_id, owner_id):
        path = f"/atlas_engine/api/v1/user_tasks/{user_task_instance_id}/reserve"

        payload = {
            'actualOwnerId': owner_id
        }

        result = await self.do_put(path, payload)

        return result

    async def cancel_reservation_user_task(self, user_task_instance_id):
        path = f"/atlas_engine/api/v1/user_tasks/{user_task_instance_id}/cancel-reservation"

        result = await self.do_put(path, {})

        return result

    async def finish_user_task(self, user_task_instance_id, answer):
        path = f"/atlas_engine/api/v1/user_tasks/{user_task_instance_id}/finish"

        result = await self.do_put(path, answer)

        return result

    async def get_user_tasks(self, state='suspended'):

        def map_user_tasks(user_task):
            new_user_task = {}
            new_user_task['userTaskInstanceId'] = user_task['flowNodeInstanceId']
            new_user_task['flowNodeId'] = user_task['flowNodeId']
            new_user_task['flowNodeName'] = user_task['flowNodeName']
            new_user_task['processModelId'] = user_task['processModelId']
            new_user_task['processInstanceId'] = user_task['processInstanceId']
            new_user_task['correlationId'] = user_task['correlationId']
            new_user_task['formFields'] = user_task['userTaskConfig']['formFields']
            new_user_task['token'] = user_task['tokens'][1]
            new_user_task['meta_info'] = user_task['metaInfo']

            # only for debug 
            #new_user_task['origin_user_task'] = user_task

            return new_user_task

        bpmn_type = 'bpmn:UserTask'
        path = f"/atlas_engine/api/v1/flow_node_instances?flowNodeType={bpmn_type}&state={state}"

        result = await self.do_get(path)

        user_tasks = []

        if result['totalCount'] > 0:
            user_tasks = list(map(map_user_tasks, result['flowNodeInstances']))

        return user_tasks



import warnings

from .external_task_client import ExternalTaskClient


class ClientWrapper:

    def __init__(self, atlas_engine_url):
        self._atlas_engine_url = atlas_engine_url

    def subscribe_to_external_task_for_topic(self, topic, handler):
        warnings.warn("Please use 'subscribe_to_external_task_topic' instead of 'subscribe_to_external_task_for_topic'.", DeprecationWarning)
        return self.subscribe_to_external_task_topic(topic, handler)

    def subscribe_to_external_task_topic(self, topic, handler):
        external_task_client = ExternalTaskClient(self._atlas_engine_url)

        external_task_client.subscribe_to_external_task_for_topic(topic, handler)

        return external_task_client
        




class FunctionalError(Exception):
    
    def __init__(self, code, message):
        super(FunctionalError, self).__init__(message)
        self._code = code
        self._message = message

    def get_code(self):
        return self._code

    def get_message(self):
        return self._message



import asyncio
import logging 

from ..core import base_client
from .start_callback_type import StartCallbackType

logger = logging.getLogger(__name__)

class ProcessDefinitionClient(base_client.BaseClient):

    def __init__(self, url, session=None, identity=None):
        super(ProcessDefinitionClient, self).__init__(url, session, identity)

    async def __start_process_instance(self, process_model_id, **options):
        start_callback_type = options.get('start_callback', StartCallbackType.ON_PROCESSINSTANCE_CREATED)
        path = f"/atlas_engine/api/v1/process_models/{process_model_id}/start"

        initial_token = options.get('inital_token', {})

        payload = {
            "returnOn": int(start_callback_type),
            "processModelId": process_model_id,
            "initialToken": initial_token
        }

        logger.info(f"start process with uri '{path}' and payload '{payload}'")

        result = await self.do_post(path, payload)

        return result

    async def __start_process_instance_by_start_event(self, process_model_id, start_event_id, end_event_id, **options):
        start_callback_type = options.get('start_callback', StartCallbackType.ON_PROCESSINSTANCE_CREATED)
        path = f"/atlas_engine/api/v1/process_models/{process_model_id}/start"

        initial_token = options.get('initial_token', {})

        payload = {
            "returnOn": int(start_callback_type),
            "processModelId": process_model_id,
            "startEventId": start_event_id,
            "initialToken": initial_token
        }

        if end_event_id is not None:
            payload["endEventId"] = end_event_id

        logger.info(f"start process with uri '{path}' and payload '{payload}'")

        result = await self.do_post(path, payload)

        return result

    def __start_process_instance_wrapper(self, process_model_id, start_event_id=None, end_event_id=None, **options):

        async def run_loop(process_model_id, start_event_id, **options):
            result = None
            
            if start_event_id is not None:
                result = await self.__start_process_instance_by_start_event(process_model_id, start_event_id, end_event_id, **options)
            else:
                result = await self.__start_process_instance(process_model_id, **options)

            await self.close()

            return result

        logger.info(f"Connection to process engine at url '{self._url}'.")
        logger.info(f"Starting process instance process_model_id '{process_model_id}' and start_event_id '{start_event_id}'.")

        loop = asyncio.new_event_loop()

        task = run_loop(process_model_id, start_event_id, **options)
        result = loop.run_until_complete(task)

        loop.close()

        return result

    def start_process_instance(self, process_model_id, start_event_id=None, **options):
        options['start_callback'] = StartCallbackType.ON_PROCESSINSTANCE_CREATED

        return self.__start_process_instance_wrapper(process_model_id, start_event_id, **options)

    def start_process_instance_and_await_end_event(self, process_model_id, start_event_id=None, **options):
        options['start_callback'] = StartCallbackType.ON_PROCESSINSTANCE_FINISHED

        return self.__start_process_instance_wrapper(process_model_id, start_event_id, **options)

    def start_process_instance_and_await_specific_end_event(self, process_model_id, start_event_id=None, end_event_id=None, **options):
        options['start_callback'] = StartCallbackType.ON_ENDEVENT_REACHED

        return self.__start_process_instance_wrapper(process_model_id, start_event_id, end_event_id, **options)

    def get_process_definition(self, process_model_id):

        async def run_loop(process_model_id):
            url = f"/atlas_engine/api/v1/process_models/{process_model_id}/process_definition"


            result = await self.do_get(url)

            return result

        logger.info(f"Get the definition of process_model '{process_model_id}' from atlas engine at url '{self._url}'.")

        loop = asyncio.new_event_loop()

        task = run_loop(process_model_id)
        result = loop.run_until_complete(task)

        loop.close()

        return result



import asyncio
from concurrent import futures
import errno
import logging
import uuid
import warnings

from ..core import BaseClient, LoopHelper
from .external_task_worker import ExternalTaskWorker

logger = logging.getLogger(__name__)

class ExternalTaskClient(BaseClient):

    def __init__(self, url, session=None, identity=None):
        super(ExternalTaskClient, self).__init__(url, session, identity)

        self.__loop_helper = LoopHelper(on_shutdown=self.__on_shutdown)
        self.__worker_id = str(uuid.uuid4())

        self._topic_for_handler = {}

    def subscribe_to_external_task_for_topic(self, topic, handler, **task_options):
        warnings.warn("Please use 'subscribe_to_external_task_topic' instead of 'subscribe_to_external_task_for_topic'.", DeprecationWarning)
        self.subscribe_to_external_task_topic(topic, handler, **task_options)

    def subscribe_to_external_task_topic(self, topic, handler, **task_options):
        """
        task_options:
            - max_tasks
            - long_polling_timeout
            - lock_duration
            - additional_duration
            - extend_lock_timeout
        """
        if topic in self._topic_for_handler:
            logger.warning(f"The topic '{topic}' skipped, is already registered.'")
        else:
            self.__start_subscription(topic, handler, task_options)

    def __start_subscription(self, topic, handler, task_options):
        async def bg_job_handle_external_tasks():
            try:
                await self.handle_external_tasks(topic, handler, task_options)
            except Exception as e:
                if type(e) is errno.EPIPE:
                    # Raised while long polling finished, that the reason for only log as info.
                    logger.info(f"expected long pollng error: {e}")
                else:
                    import traceback
                    import sys
                    traceback.print_exc(file=sys.stdout)
                    logger.warn(f"exception on handle_external_task: {e}")

        async_bg_task = self.__loop_helper.register_background_task(bg_job_handle_external_tasks)
        self._topic_for_handler[topic] = async_bg_task

    def subscribe_to_external_tasks_for_topics(self, topics_for_handlers):
        for topic in topics_for_handlers.keys():
            handler = topics_for_handlers[topic] # TODO: Options sollten auch hier zu setzen sein

            self.subscribe_to_external_task_for_topic(topic, handler)

    def start(self):
        logger.info(f"Starting external task for topics '{', '.join(self._topic_for_handler.keys())}'.")
        logger.info(f"Connecting to process engine at url '{self._url}'.")
        self.__loop_helper.start()

    async def handle_external_tasks(self, topic, handler, task_options):
        external_tasks = await self.__fetch_and_lock(topic, handler, task_options)

        len_external_tasks = len(external_tasks)
        if len_external_tasks >= 1:
            logger.info(f"receive {len_external_tasks} tasks for topic '{topic}'.")

            external_task_tasks = []

            external_task_options = {}

            if task_options.get("additional_lock_duration", None):
                external_task_options["additional_lock_duration"] = task_options["additional_lock_duration"]

            # EXTEND_LOCK_TIMEOUT
            if task_options.get("extend_lock_timeout", None):
                external_task_options["extend_lock_timeout"] = task_options["extend_lock_timeout"]

            for external_task in external_tasks:
                external_task_worker = ExternalTaskWorker(
                    self._url, 
                    self._session, 
                    self._identity, 
                    self.__loop_helper,
                    handler, 
                    external_task,
                    external_task_options)
                task = external_task_worker.start()
                external_task_tasks.append(task)

            try:
                done_tasks, pending = await asyncio.wait(external_task_tasks, return_when=futures.ALL_COMPLETED)
                logger.debug(f"Wait result while handle external task (done: {done_tasks}, pending {pending}).")
            except Exception as e:
                logger.error(f"asyncio.wait({e} - {type(e)})")

    async def __on_shutdown(self):
        await self.close()

    def stop(self):
        logger.info(f"Stopping external task for topics {self._topic_for_handler.keys()}")
        self.__loop_helper.stop()

    async def __fetch_and_lock(self, topic, handler, task_options={}):
        logger.info(f"fetch and lock external task for topic {topic} and options {task_options}")

        max_tasks = task_options.get('max_tasks', 10)
        long_polling_timeout = task_options.get('long_polling_timeout', (10 * 1000))
        lock_duration = task_options.get('lock_duration', (100 * 1000))

        request = {
            "workerId": self.__worker_id,
            "topicName": topic,
            "maxTasks": max_tasks,
            "longPollingTimeout": long_polling_timeout,
            "lockDuration": lock_duration
        }

        external_tasks = await self.do_post('/atlas_engine/api/v1/external_tasks/fetch_and_lock', request)

        logger.debug(f"receive {len(external_tasks)} for topic {topic}.")

        return external_tasks



import asyncio
import logging 

from ..core import BaseClient, LoopHelper

logger = logging.getLogger(__name__)

class NotificationClient(BaseClient):
    def __init__(self, url, session=None, identity=None):
        super(NotificationClient, self).__init__(url, session, identity)

        self.__loop_helper = LoopHelper(on_shutdown=self.__on_shutdown)
        self.__notification_callbacks = {}

    async def __on_shutdown(self):
        for task in self.__notification_callbacks.values():
            self.__loop_helper.unregister_background_task(task)

        await self.close()

    def start(self):
        logger.info(f"Connecting to process engine at url '{self._url}'.")
        self.__loop_helper.start()

    def __on_register_event(self, event_name, callback):

        async def bg_on_event():

            url = f"/atlas_engine/api/v1/notifications/long_polling/{event_name}"

            result = await self.do_get(url)

            logger.debug(f"handling event '{event_name}' with {result}")

            if asyncio.iscoroutine(callback):
                await callback(result)
            else:
                callback(result)

        async_bg_task = self.__loop_helper.register_background_task(bg_on_event)
        self.__notification_callbacks[event_name] = async_bg_task

    def on_activity_reached(self, callback):
        self.__on_register_event('activity_reached', callback)

    def on_activity_finished(self, callback):
        self.__on_register_event('activity_finished', callback)

    def on_boundary_event_triggered(self, callback):
        self.__on_register_event('boundary_event_triggered', callback)

    def on_empty_activity_waiting(self, callback):
        self.__on_register_event('empty_activity_waiting', callback)

    def on_empty_activity_finished(self, callback):
        self.__on_register_event('empty_activity_finished', callback)

    def on_manual_task_waiting(self, callback):
        self.__on_register_event('manual_task_waiting', callback)

    def on_manual_task_finished(self, callback):
        self.__on_register_event('manual_task_finished', callback)

    def on_process_started(self, callback):
        self.__on_register_event('process_started', callback)

    def on_process_ended(self, callback):
        self.__on_register_event('process_ended', callback)

    def on_process_error(self, callback):
        self.__on_register_event('process_error', callback)

    def on_user_task_waiting(self, callback):
        self.__on_register_event('user_task_waiting', callback)

    def on_user_task_finished(self, callback):
        self.__on_register_event('user_task_finished', callback)

    def on_empty_activity_finished(self, callback):
        self.__on_register_event('empty_activity_finished', callback)

    def on_empty_activity_waiting(self, callback):
        self.__on_register_event('empty_activity_waiting', callback)

    def on_manual_task_finished(self, callback):
        self.__on_register_event('manual_task_finished', callback)

    def on_manual_task_waiting(self, callback):
        self.__on_register_event('manual_task_waiting', callback)

    def on_intermediate_throw_event_triggered(self, callback):
        self.__on_register_event('intermediate_throw_event_triggered', callback)

    def on_intermediate_catch_event_reached(self, callback):
        self.__on_register_event('intermediate_catch_event_reached', callback)

    def on_intermediate_catch_event_finished(self, callback):
        self.__on_register_event('intermediate_catch_event_finished', callback)




import asyncio
import logging 

from ..core import base_client

logger = logging.getLogger(__name__)

class FlowNodeInstanceClient(base_client.BaseClient):

    def __init__(self, url, session=None, identity=None):
        super(FlowNodeInstanceClient, self).__init__(url, session, identity)

    async def __trigger_message_event(self, event_name, process_instance_id=None):
        
        url = f"/atlas_engine/api/v1/messages/{event_name}/trigger"

        if process_instance_id is not None:
            url = f"{url}?processInstanceId={process_instance_id}"

        result = await self.do_post(url, {})

        return result

    def trigger_message_event(self, event_name, process_instance_id=None):

        async def run_loop(event_name, process_instance_id):

            result = await self.__trigger_message_event(event_name, process_instance_id)

            return result

        logger.info(f"Connection to process engine at url '{self._url}'.")
        logger.info(f"Trigger message '{event_name}'.")

        loop = asyncio.new_event_loop()

        task = run_loop(event_name, process_instance_id)
        result = loop.run_until_complete(task)

        loop.close()

        return result

    async def __trigger_signal_event(self, event_name):

        url = f"/atlas_engine/api/v1/signals/{event_name}/trigger"

        result = await self.do_post(url, {})

        return result

    def trigger_signal_event(self, event_name):

        async def run_loop(event_name):
            result = await self.__trigger_signal_event(event_name)

            return result

        logger.info(f"Connection to process engine at url '{self._url}'.")
        logger.info(f"Trigger signal '{event_name}'.")

        loop = asyncio.new_event_loop()

        task = run_loop(event_name)
        result = loop.run_until_complete(task)

        loop.close()

        return result


import setuptools

# TODO: mm - requirements.txt verwenden - https://stackoverflow.com/questions/14399534/reference-requirements-txt-for-the-install-requires-kwarg-in-setuptools-setup-py
# TODO: mm - twine benutzen https://pypi.org/project/twine/
#
# Kurzanleitung
#    https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='atlas_engine_client',
    version=setuptools.sic('1.0.0-beta.1'),
    author="5Minds IT-Solutions GmbH & Co. KG",
    author_email="atlasengine@5minds.de",
    description="A Client for an atlas-engine.io hosted workflow engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="workflow-engine atlas-engine client bpmn",
    url="https://github.com/atlas-engine/Client.py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
